package edu.kit.iti.formal.lights;

import org.jetbrains.annotations.NotNull;
import org.sat4j.minisat.SolverFactory;
import org.sat4j.specs.ContradictionException;
import org.sat4j.specs.ISolver;
import org.sat4j.specs.TimeoutException;

/**
 * An spotlight solver is an object that allows you to solve spotlight puzzles.
 * <p>
 * It has one central method which takes a board description and returns a
 * solution for this board.
 *
 * @author Mattias Ulbrich
 * @author Alexander Weigl
 * @version 2019.0
 */
public abstract class LightUpSolver {
    /**
     * The Lights object we use.
     */
    protected final Lights lights;
    /**
     * An instance of the SAT solver.
     */
    protected ISolver solver = SolverFactory.newDefault();
    /**
     * The dimension, taken from {@link #lights}.
     */
    protected int dimension;

    /**
     * Store your solution here.
     */
    protected final Solution solution;

    protected LightUpSolver(@NotNull Lights lights) {
        this.lights = lights;
        dimension = lights.getDimension();
        solution = new Solution(dimension);
    }


    /**
     * This method computes the solution for a lightUp puzzle.
     *
     * <p>
     * The implementation must take a non-<code>null</code> String describing a
     * board. The board description is in the format accepted by
     * {@link Lights#Lights(String)}.
     *
     * <p>
     * It must also return a String which is in the format returned by
     * {@link Lights#toString()}. and return a String. May return null to
     * indicate that no solution is available.
     *
     * <p>
     * If no solution can be found, null can be returned.
     *
     * @param lights the board , not <code>null</code>
     * @return a board description or <code>null</code>
     * @throws LightsFormatException if desc is not correctly formatted.
     * @throws TimeoutException      if the sat solver times out.
     */
    public abstract Solution solve() throws LightsFormatException, TimeoutException, ContradictionException;
}
