package edu.kit.iti.formal.lights;

import org.sat4j.specs.ContradictionException;
import org.sat4j.specs.TimeoutException;

import java.awt.*;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.NoSuchElementException;
import java.util.Optional;

/**
 * @author Alexander Weigl
 * @version 22.11.2016, 09.01.2016, 03.09.19, 05.11.2019
 */
public class TestRunner implements Runnable {
    /**
     * If this flag is true, then testing a found solution is skipped.
     */
    public static final boolean NO_CHECK =
            Boolean.getBoolean("runner.nocheck");


    /**
     * If this flag is true, then SVG files are created for every game instance.
     */
    public static final boolean SVG =
            Boolean.getBoolean("runner.svg");

    /**
     * Folder with riddles instances.
     */
    public static final String RIDDLES_FOLDER = System.getenv().
            getOrDefault("runner.riddles", "./riddles/");
    /**
     * The timeout of the runner in seconds
     */
    private static final int TIMEOUT_IN_SECONDS =
            Integer.getInteger("runner.timeout", 300);

    /**
     *
     */
    public static final String CLASSNAME =
            System.getProperty("runner.classname", "MyLightUpSolver");


    //statistics
    private ArrayList<Long> timings = new ArrayList<>(100);
    private int success, all;
    private long startTime, stopTime;

    private static File[] findRiddles() {
        final File folder = new File(RIDDLES_FOLDER);
        System.out.format("I am using the riddle folder '%s'\n", folder);
        if (!folder.exists()) {
            System.err.format("Folder %s does not exist. ABORT!\n", folder);
            System.exit(1);
        }
        File[] riddles = folder.listFiles();
        if (riddles.length == 0) {
            System.err.format("No riddles in folder %s found. ABORT!\n", folder);
            System.exit(1);
        } else {
            System.out.format("Found %d riddles!\n", riddles.length);
        }

        return riddles;
    }

    public static LightUpSolver solverInstance(Lights puzzle)
            throws NoSuchMethodException, ClassNotFoundException,
            IllegalAccessException, InvocationTargetException,
            InstantiationException {
        Class<LightUpSolver> clazz = (Class<LightUpSolver>) Class.forName(CLASSNAME);
        try {
            return clazz.getConstructor(Lights.class).newInstance(puzzle);
        } catch (InstantiationException | NoSuchMethodException e) {
            for (Constructor<?> c : clazz.getConstructors()) {
                System.out.println("Constructor found: " + c);
            }
            throw e;
        }
    }

    public static void main(String[] args) {
        TestRunner test = new TestRunner();
        Thread t = new Thread(test);
        t.start();
        try {
            t.join(TIMEOUT_IN_SECONDS * 1000);
        } catch (InterruptedException e) {
            //
        } finally {
            test.statistics();
            System.exit(0);
        }
    }

    public boolean checkSolution(File file) throws IOException,
            ClassNotFoundException, NoSuchMethodException, InvocationTargetException, InstantiationException,
            IllegalAccessException, LightsFormatException, TimeoutException, ContradictionException {
        Lights game = Lights.fromFile(file);
        System.out.printf("== %s ======================================================%n", file.getName());
        LightUpSolver solver = solverInstance(game);

        long start = System.currentTimeMillis();
        Solution b = solver.solve();
        timings.add(System.currentTimeMillis() - start);

        final boolean expectedSat = file.getName().startsWith("sat");
        final boolean actualSat = b != null;

        System.out.println("Riddle is " + (!expectedSat ? "UNSAT " : "SAT"));
        System.out.println("Your answer is " + (actualSat ? "SAT" : "UNSAT"));

        boolean solved = actualSat == expectedSat;
        if (actualSat && !NO_CHECK) {
            var sc = new SolutionChecker(game, b);
            if (SVG) {
                SVGWriter.writeFile(file.getName() + ".svg", game, sc);
                System.out.format("Write svg file '%s' for game instance '%s'\n",
                        file.getName() + ".svg", file);
            }

            if (!sc.isValid()) {
                System.out.print("Your solution contains errors:\n");
                for (String error : sc.getErrors()) {
                    System.out.println("\t" + error);
                }
            }
            solved &= sc.isValid();
        }

        if (solved) System.out.println("SOLVED!");
        else System.out.println("NOT SOLVED :(");
        return solved;
    }

    public void run() {
        System.out.println("SVG is " + (SVG ? "on" : "off"));
        System.out.println("Solution Check is " + (!NO_CHECK ? "on" : "off"));
        System.out.println("Timeout is " + TIMEOUT_IN_SECONDS);

        try {
            startTime = System.currentTimeMillis();
            success = 0;
            all = 0;
            for (File file : findRiddles()) {
                all++;
                try {
                    success += checkSolution(file) ? 1 : 0;
                } catch (IOException e) {
                    System.err.println("Error while loading " + file);
                } catch (IllegalArgumentException | NoSuchElementException | HeadlessException
                        | LightsFormatException | TimeoutException | ContradictionException e) {
                    e.printStackTrace();
                }
            }
            stopTime = System.currentTimeMillis();
        } catch (ClassNotFoundException e) {
            System.err.println("Could not load the solution by class name.");
            System.err.println("There should be class 'MyLightUpSolver' in the default package on the classpath.");
            e.printStackTrace();
        } catch (NoSuchMethodException | InvocationTargetException | InstantiationException | IllegalAccessException e) {
            System.err.println("Constructor not found");
            System.err.println("There should be constructor 'public MyLightUpSolver(Lights)' in your solution class.");
            e.printStackTrace();
        }
    }

    private void statistics() {
        System.out.format("%%%% %s: Score %d of %d (%2.1f%%) successful!%n",
                new File(".").getAbsolutePath(),
                success, all, 100.0 * success / all);
        System.out.printf("%%%% %s: %10d ms%n",
                new File(".").getAbsolutePath(),
                stopTime - startTime);

        Optional<Long> optMax = timings.stream().max(Long::compare);
        Optional<Long> optMin = timings.stream().min(Long::compare);

        if (optMax.isPresent() && optMin.isPresent()) {
            long max = optMax.get();
            long min = optMin.get();
            long sum = timings.stream().reduce(Long::sum).get();
            System.out.printf("%% Max time: %10d ms%n", max);
            System.out.printf("%% Min time: %10d ms%n", min);
            System.out.printf("%% Avg time: %10.0f ms%n", sum / (float) timings.size());
        }
        System.out.println("%% end");
    }
}
